/*++++++++++++++
 Copyright:    (C) 2008-2015 UDS/CNRS
 License:       GNU General Public License
.IDENTIFICATION blook.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Bib datafiles
.KEYWORDS       Index based on Paragraphs
.VERSION  1.0   04-Feb-1998
.VERSION  1.1   23-Mar-1999: Several extensions
.VERSION  1.2   23-Jun-1999: -d -f -TeX
.VERSION  1.3   29-Oct-2000: In -d option, be sure punctuation found
.VERSION  1.4   27-Feb-2002: Change the gsearch alogrithm; added diff1_tex
.VERSION  1.5   17-Jan-2010: Cosmetics
.COMMENTS       From a set of input files which are sorted,
	write out the file contents which match the user's input.
	This program also assumes bibcodes with -bib option
---------------*/

static char usage[] = "\
Usage: blook [-a[#]] [-d|-f|-TeX] [-t[#]] [-bib] [.ext] sorted_file...\n\
" ;
static char help[] = "\
    -a: appends input to each line, eventually preceded by # (default tab)\n\
  -bib: the input consists of bibcodes, which is skipped in the output\n\
	When several sorted_files are provided, it is assumed that the\n\
	     sorted_file to choose is terminaed with the journal code\n\
    -d: dictionary order (as in sort)\n\
    -f: ignore case (case-insensitive input)\n\
    -T: TeX macros ignored\n\
    -t: issue only text FOLLOWING the specified separator char (default tab)\n\
  .ext: the specified extension must be added to sorted_file to get the index\n\
	     (until a new .ext)\n\
  sorted_file is the name of files which exist.\n\
" ;

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif

		/* In MSDOS, lines are terminated by \r\n ... */
#ifdef __MSDOS__
# ifndef INT4
# define INT4		long
# endif
# define BUFSIZE	4096
# define MAXFILES	25	/* Maximal Number of Files */
# define htonl(x)	(x)	/* No byte swap	*/
# define ntohl(x)	(x)	/* No byte swap	*/
# define gethostid()	0
# define getpid()	1
static char *tmp_dir  = "\\tmp\\par";
# define EOL		'\r'
  static char *delim="\r\n";		/* Delimiter expr. */
#else
# ifndef INT4
# define INT4		int
# endif
# include <netinet/in.h>	/* hton (Host to Net)	   */
# define BUFSIZE	8192
# define MAXFILES	25	/* Maximal Number of Files */
# define gethostid()	time(0)
static char *tmp_dir  = "/tmp/";
# define EOL		'\n'
  static char *delim="\n";		/* Delimiter expr. */
#endif


#define bcopy(s,d,l)	memcpy(d,s,l)
#define ITEMS(s)	(sizeof(s)/sizeof(s[0]))
#define MIN(a,b)	( a <= b ? a : b)
#define MAX(a,b)	( a >= b ? a : b)

typedef struct Sfile {	/* Sorted file */
	char *name ;
	char *text ;	/* Contents of file */
	int a, u, o;	/* Sizes of buffer  */
	short suflen ;	/* Length of suffix */
	char obib ;	/* Current obib     */
} SFILE ;
static SFILE sfile[MAXFILES] ;

static int nfiles;	/* Number of files  */
static char *ext;	/* Extension */
static int (*fdiff)() ;	/* Comparison funct */
static int (*fdiff1)();	/* Comparison funct */

static char buf[BUFSIZE];
static char ov  = 0;	/* Verbose Option  */
static char optd  = 0;	/*    -d   Option  */
static char obib= 0;	/* -bib   option   */
static char sep_char ;	/* -t option	   */
static char *atxt ;
static int matched_length ; /* set by dif_ */

/* Macros with their interpretation; default = text */
char *tex_symb[] = {
  "AA", "A",
  "aa", "a",
  "c", "",      /* Accent */
  "v", "",      /* Accent */
  "deg", "",    /* Accent */
  "&", "&",
  "$", "$",
  (char *)0
} ;


/*==================================================================
		General Utility Routines
 *==================================================================*/

#ifdef MISS_strncasecmp
static int strncasecmp(s1, s2, len)
/*++++++++++++++++
.PURPOSE  Compare two strings, case insensitive
.RETURNS  The difference
-----------------*/
  char *s1 ;	/* IN: First string 	*/
  char *s2 ;	/* IN: Second string 	*/
  int len ;	/* IN: Length of Comparison */
{
  char *e1;
  int diff ;
    e1 = s1 + len ;
    diff = 0; 
    while ((diff == 0) && (s1 < e1)) 
	diff = toupper(*s1) - toupper(*s2) ;
    return(diff) ;
}
#endif

static char *strdup2(s1, s2)
/*++++++++++++++++
.PURPOSE  Create a new string by concatenation
.RETURNS  The concatenated (new) string
-----------------*/
  char *s1 ;	/* IN: First string 	*/
  char *s2 ;	/* IN: Second string 	*/
{
  char *s; int len ;
    len = strlen(s1) + 1 ;
    if (s2) len += strlen(s2) ;
    if ((s = malloc(len))) {
	strcpy(s, s1);
	if (s2) strcat(s, s2) ;
    }
    return(s) ;
}

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

static int strbloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character backward
.RETURNS  Index of located char / -1 if not found
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for (s=text+strlen(text); --s >= text; )  if (*s == c)    break;
    return(s-text);
}

static char *Gets()
/*++++++++++++++++
.PURPOSE  Get a line (and forget the end of) on stdin
.RETURNS  as gets
-----------------*/
{
  static char *abuf; static int lbuf;
  int i;
    
    if (!lbuf) { lbuf = BUFSIZ-1; abuf = malloc(lbuf+1); abuf[lbuf] = 0; }

    if (isatty(0)) 
	fprintf(stderr, "----Give a %s: ", obib ? "bibcode" : "Key") ;
    if (! fgets(abuf, lbuf-1, stdin)) {	/* Error or EOF ? */
	if (ferror(stdin)) perror("****Gets");
	return((char *)0);
    }
    i = strloc(abuf, '\n');
    while (!abuf[i]) {
	lbuf += BUFSIZ; abuf = realloc(abuf, lbuf);
	if (!fgets(abuf+i, BUFSIZ, stdin)) break;
	i = strloc(abuf, '\n');
    }
    while ((i>=0) && ((abuf[i] == '\n') || (abuf[i] == '\r'))) i-- ;
    abuf[++i] = 0;
    return(abuf);
}

static char *lookup(dic, term)
/*++++++++++++++++
.PURPOSE  Find equivalence in a Dictionary
.RETURNS  The equivqlence : NULL
-----------------*/
  char **dic ;	/* IN: The Dictionary	*/
  char *term ;	/* IN: The term lookup	*/
{
  char *p, *t ;
    while (*dic) {
	for (p = *dic, t = term; isalpha(*p) && isalpha(*t) && (*p == *t); 
	    p++, t++) ;
	if ((!isalpha(*p)) && (!isalpha(*t))) return(dic[1]) ;
        dic += 2 ;
    }
    return ((char *)0) ;
}

/*==================================================================
		Comparison (difference) Functions
 *==================================================================*/
static int dif_d(str1, str2, len)
/*++++++++++++++++
.PURPOSE  Compare two strings, dictionary order, maximum len bytes
.RETURNS  The difference (0 when matching)
.REMARKS  len applies to s1.
	In case of match, the static matched_length of s2 
-----------------*/
  char *str1;	/* IN: First string	*/
  char *str2;	/* IN: Second string	*/
  int len ;	/* IN: Max. str1 length	*/
{
  char *e1, *s1, *s2;
  int diff ;
    s1 = str1 ; s2 = str2 ;
    e1 = s1 + len ;
    diff = 0; 
    while ((diff == 0) && (s1 < e1)) {
	if (ispunct(*s1)) { s1++; continue; }
	if (ispunct(*s2)) { s2++; continue; }
	diff = toupper(*s1) - toupper(*s2) ;
	if (isspace(*s1)) while (isspace(*s1)) s1++ ;
	else s1++;
	if (isspace(*s2)) while (isspace(*s2)) s2++ ;
	else s2++;
    }
    matched_length = s2 - str2 ;
    return(diff) ;
}

static int skip_tex(str, equiv)
/*++++++++++++++++
.PURPOSE  Compare two strings, latex order, case inensitive, maximum len bytes
.RETURNS  The length of macro to skip
-----------------*/
  char *str;    /* IN: string to check, starting by \\  */
  char *equiv;  /* OUT: Equivalence of Input */
{
  char *s, *s0, *eq ; int m ;
    s = str ;
    equiv[0] = 0 ;
    if (*s == '\\') s++ ;
    if (isalpha(*s)) {
        s0 = s ;
        while (isalpha(*s)) s++ ;
        if ((eq = lookup(tex_symb, s0))) strcpy(equiv, eq) ;
        else {
            strncpy(equiv, s0, m = s - s0) ;
            equiv[m] = 0;
        }
        while (*s == ' ') s++ ;
    }
    else /* e.g. \'e */ s++ ;
    return(s-str) ;
}

static int dif_tex(str1, str2, len)
/*++++++++++++++++
.PURPOSE  Compare two strings, latex order, case inensitive, maximum len bytes
.RETURNS  The difference (0 when matching)
.REMARKS  len applies to s1
-----------------*/
  char *str1;	/* IN: First string     */
  char *str2;	/* IN: Second string    */
  int len ;     /* IN: Max.str1 length  */
{
  static char *n1, *n2 ;
  static int recu ;
  char *e1, *s1, *s2, *sn1, *sn2, eq1[20], eq2[20];
  int diff ;
    s1 = str1 ; s2 = str2 ;
    e1 = s1 + len ;
    diff = 0;
    recu++;
    while ((diff == 0) && (s1 < e1)) {
        if (*s1 == '{') { s1++; continue; }
        else if (*s1 == '}') { s1++; continue; }
        else if (*s1 == '\\') {
            s1 += skip_tex(s1, eq1);
            if (eq1[0]) {
                sn1 = n1; sn2 = n2 ;
                diff = dif_tex(eq1, s2, strlen(eq1)) ;
                s2 = n2 ;
                n1 = sn1 ; n2 = sn2 ;
            }
            continue;
        }
        else if (optd && ispunct(*s1)) { s1++; continue; }
        if (*s2 == '{') { s2++; continue; }
        else if (*s2 == '}') { s2++; continue; }
        else if (*s2 == '\\') {
            s2 += skip_tex(s2, eq2);
            if (eq2[0]) {
                sn1 = n1; sn2 = n2 ;
                diff = dif_tex(s1, eq2, strlen(eq2)) ;
                s1 = n1 ;
                n1 = sn1 ; n2 = sn2 ;
            }
            continue;
        }
        else if (optd && ispunct(*s2)) { s2++; continue; }
        diff = toupper(*s1) - toupper(*s2) ;
        if (isspace(*s1)) while (isspace(*s1)) s1++ ;
        else s1++ ;
        if (isspace(*s2)) while (isspace(*s2)) s2++ ;
        else s2++ ;
    }
    n1 = s1 ; n2 = s2 ; /* Save the new new position */
    recu-- ;
    matched_length = (s2 - str2) ;
    return(diff) ;
}

static int dif1_tex(str1, str2, len)
/*++++++++++++++++
.PURPOSE  Compare two latex strings, case insensitive -- punctuations must match
.RETURNS  The difference (0 when matching)
.REMARKS  len applies to s1
-----------------*/
  char *str1;	/* IN: First string     */
  char *str2;	/* IN: Second string    */
  int len ;     /* IN: Max.str1 length  */
{
  static char *n1, *n2 ;
  static int recu ;
  char *e1, *s1, *s2;
  int diff ;
    s1 = str1 ; s2 = str2 ;
    e1 = s1 + len ;
    diff = 0;
    recu++;
    while ((diff == 0) && (s1 < e1)) {
        if (*s1 == '{' || *s1 == '}' || *s1 == '\\') { s1++; continue; }
        if (*s2 == '{' || *s2 == '}' || *s2 == '\\') { s2++; continue; }
        diff = toupper(*s1) - toupper(*s2) ;
        if (isspace(*s1)) while (isspace(*s1)) s1++ ;
        else s1++ ;
        if (isspace(*s2)) while (isspace(*s2)) s2++ ;
        else s2++ ;
    }
    matched_length = (s2 - str2) ;
    return(diff) ;
}


/*==================================================================
		Dealing with Index
 *==================================================================*/

static char *gsearch(buf, len, key)
/*++++++++++++++++
.PURPOSE  Search in buf line starting with key
.RETURNS  The starting line where   line <= key <= (next line)
.REMARKS  buf[-1] assumed to be a newline
-----------------*/
  char *buf;	/* IN: Buffer with lines     */
  int  len;	/* IN: Length of buf	     */
  char *key;	/* IN: String to find in buf */
{
  char *p0, *p1, *p ;
  int lk;

    lk = strlen(key) ;
    p0=buf, p1=p0+len ;
    if ((*fdiff)(key, p0, lk) > 0) while (p0<p1) {
    	/* if ((*fdiff)(key, p0, lk) <= 0) break; */
    	p = p0 + (p1-p0)/2;
    	while (*p != '\n') p-- ; ++p;
	if (p == p0) {
	    while (*p != '\n') p++; 
	    ++p;
	}
    	if ((*fdiff)(key, p, lk) > 0) p0 = p;
    	else {
	    if (p1 == p) p0 = p ;
	    else p1 = p;
	}
    }
    return(p0);
}

/*==================================================================
		Display File's Extracts
 *==================================================================*/

static int LoadFile(sf)
/*++++++++++++++++
.PURPOSE  Load a file into memory
.RETURNS  -1 = error
.REMARKS  
-----------------*/
  SFILE *sf ;	/* MOD: File to load   */
{
  int len, i;
    
    i = open(sf->name, O_BINARY) ;
    if (i <= 0) {
	fprintf(stderr, "****") ;
	perror(sf->name) ;
	return(i) ;
    }

    sf->u = lseek(i, 0L, 2); lseek(i, 0L, 0);
    sf->a = 1+(sf->u|15) ;
    sf->text = malloc(sf->a) ;
    sf->text[0] = '\n' ;	/* First line always empty */
    len = read(i, sf->text+1, sf->u) ;
    if (len != sf->u) {
	fprintf(stderr, "****Error reading %s (%d instead of %d bytes\n", 
	  sf->name, len, sf->u) ;
	return(-1) ;
    }
    if (len < 100) fprintf(stderr, 
	"++++File %s is only %d bytes\n", sf->name, len) ;
    sf->u += 1 ;
    sf->text[sf->u] = 0 ;
    close(i);
    return(0);
}

static int DisplayKey(sfno, str)
/*++++++++++++++++
.PURPOSE  Display the line matching the input string
.RETURNS  Number of lines found
.REMARKS  Output to stdout.
-----------------*/
  int  sfno;	/* IN: SFILE number    */
  char *str;	/* IN: The key to find */
{
  SFILE *sf ; char *p;
  int punct, opunct, len, i, n ;

    /* Load the file in memory */
    sf = &sfile[sfno] ;
    if (!sf->text) {
	i = LoadFile(sf) ;
	if (i<0) return(0) ;
    }

    /* When special character specfied... */
    punct = opunct = 0 ;
    if (optd) for (p = str; *p && (!punct); p++) {
	if (ispunct(*p)) punct = *p, opunct = (p-str) ;
    }

    n = 0; 
    len = strlen(str) ;
    if ((p = gsearch(sf->text, sf->u, str))) {
	while ((*fdiff)(str, p, len) == 0) {
	    /* dif_xx routine save in matched_length #bytes matched in p */
#if 1	    /* Version Feb. 2002 */
	    if (fdiff != fdiff1) {
		if ((*fdiff1)(str, p, len)) {
		    p += strloc(p, '\n') ;
		    if (*p == '\n') p++ ;
		    continue ;
		}
	    }
#else
	    if (punct) { 
		s = p ;
		while ((*p != punct) && (*p != '\n') && *p) p++ ;
		if ((*p != punct) || ((p-s) >= matched_length)) {
		    if (*p == '\n') p++ ;
		    continue ;
		}
		p = s ;
	    }
#endif
	    n++ ;
	    if (sep_char) {
		while ((*p != sep_char) && (*p != '\n') && *p) p++ ;
		if (*p == sep_char) p++ ;
	    }
	    else if (obib) { p += 19 ; if (*p != '\n') p++; }
	    if (nfiles > 1) printf("%d ", sfno) ;
	    while (*p != '\n') putchar(*p), p++ ;
	    if (atxt) printf("%s%s", atxt, str) ;
	    putchar('\n') ;
	    if (*p == '\n') p++ ;
	}
    }

    return(n) ;
}

/*==================================================================*/

int main(int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  to OS
.REMARKS  
-----------------*/
{
  char jnl[8], *p;
  char *line;	/* Input */
  char *ext = (char *)0 ;
  int  stobib = 0 ;
  int  i, n;

    fdiff = fdiff1 = strncmp ;	/* Default Comparison Function 	*/

    /* Take care of Options */
    while (--argc > 0) {
	p = *++argv;
	if (*p == '-') switch(p[1]) {	/* Option 	*/
	    case '-':
		stobib = 0 ;
		continue ;
	    case 't':
		sep_char = p[2] ;
		if (!sep_char) sep_char = '\t';
		continue ;
	    case 'a':
		atxt = p+2 ;
		if (!p[2]) atxt = "\t" ;
		continue ;
	    case 'b':
		stobib = obib = 1;	
		continue;
	    case 'h':
		printf("%s%s", usage, help) ;
		exit(0) ;
	    case 'd':			/* Dictionary Order */
		if (fdiff == strncmp) fdiff = dif_d ;
		if (fdiff1== strncmp) fdiff1= dif_d ;
		optd = 1 ;
		continue ;
	    case 'T':			/* Dictionary Order */
		fdiff = dif_tex ;
		fdiff1= dif1_tex ;
		continue ;
	    case 'f':			/* Case-insensitive */
		fdiff = fdiff1 = strncasecmp ;
		continue ;
	    case 'v': 			/* Verbose	*/
		ov = 1;
		continue;
	    default:
	    bad_option:
	    	fprintf(stderr, "*** blook: Invalid Option: %s\n", p);
		fprintf(stderr, "%s", usage) ;
	    	exit(1) ;
	}
	if ((*p == '.') && (!p[strloc(p,'/')])) { 
	    if (p[1] == '\0') ext = (char *)0 ;
	    else  ext = p; 
	    continue ; 
	}
	if (nfiles >= MAXFILES) {
	    fprintf(stderr, "*** blook: Too many (%d) files: %s\n", 
	    	nfiles, p);
	    fprintf(stderr, "%s", usage) ;
	    exit(1) ;
	}
	if (ext) {
	    sfile[nfiles].name = strdup2(p, ext);
	    sfile[nfiles].suflen = strlen(ext) ;
	}
	else sfile[nfiles].name = p ;
	sfile[nfiles].name = strdup2(p, ext);
	sfile[nfiles].obib = stobib ;
	nfiles++;
    }

    /* Check everything is here */
    if (!nfiles) {
	fprintf(stderr, "*** blook: Missing file(s) arguments\n");
	fprintf(stderr, "%s", usage) ;
	exit(1) ;
    }
    /* if (obib && (nfiles > 1)) obib = nfiles ; */

    while ((line = Gets())) {
	if (obib) {
	    if (!isdigit(*line)) { puts(line); continue; }
	    for (i=4; (i<9) && line[i] && (line[i] != '.'); i++) {
		if (line[i] == '+') line[i] = '&' ; 
		jnl[i-4] = line[i] == '&' ? '+' : line[i] ; 
	    }
	    jnl[i-4] = 0 ;
	    n = 0 ;
	    for (i=0; (i<nfiles) && sfile[i].obib; i++) {
		p = sfile[i].name ; p += strlen(p) ;
		p -= sfile[i].suflen + strlen(jnl) ;
		if (strncmp(jnl, p, strlen(jnl)) != 0) continue ;
		if (ispunct(p[-1])) break ;
	    }
	    /* Display from all files until something found... */
	    n = DisplayKey(i, line) ;
	    if (n == 0) {
		while ((i < nfiles) && (sfile[i].obib)) i++ ;
	    }
	    while ((n == 0) && (i < nfiles)) {
		n = DisplayKey(i, line) ;
		i++ ;
	    }
	    if (n == 0) printf ("%s%s\n", atxt ? atxt : "#", line) ;
	    continue ;
	}

	/* Not bibcode in input */
	if (*line == '#') { puts(line); continue; }

	/* Find the key in all files */
	n = 0 ;
	for (i=0; i<nfiles; i++)
	    n += DisplayKey(i, line) ;
	if ((n == 0) && atxt) printf("%s%s\n", atxt, line) ;
    }
    exit(0);
}
