/*++++++++++++++
 Copyright:    (C) 2008-2015 UDS/CNRS
 License:       GNU General Public License
.IDENTIFICATION bsort.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Bib datafiles
.KEYWORDS       Index based on Paragraphs
.VERSION  1.0   24-Jun-1999
.VERSION  1.1   17-Jan-2010: Cosmetics
.COMMENTS       Sort a file with options not in standard sort
---------------*/

static char usage[] = "\
Usage: bsort [-d|-f|-TeX] [-o output_file] file...\n\
" ;
static char help[] = "\
    -d: dictionary order (as in sort)\n\
    -f: fold case (as in sort)\n\
    -T: TeX macros ignored\n\
    -o: specifies the output file (can be identical to input file)\n\
  file is the name of files which exist.\n\
" ;

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif

		/* In MSDOS, lines are terminated by \r\n ... */
#ifdef __MSDOS__
# ifndef INT4
# define INT4		long
# endif
# define BUFSIZE	4096
# define MAXFILES	25	/* Maximal Number of Files */
# define htonl(x)	(x)	/* No byte swap	*/
# define ntohl(x)	(x)	/* No byte swap	*/
# define gethostid()	0
# define getpid()	1
static char *tmp_dir  = "\\tmp\\par";
# define EOL		'\r'
  static char *delim="\r\n";		/* Delimiter expr. */
#else
# ifndef INT4
# define INT4		int
# endif
# include <netinet/in.h>	/* hton (Host to Net)	   */
# define BUFSIZE	8192
# define MAXFILES	25	/* Maximal Number of Files */
# define gethostid()	time(0)
# define EOL		'\n'
#endif


#define bcopy(s,d,l)	memcpy(d,s,l)
#define ITEMS(s)	(sizeof(s)/sizeof(s[0]))
#define MIN(a,b)	( a <= b ? a : b)
#define MAX(a,b)	( a >= b ? a : b)

typedef struct Sfile {	/* Sorted file */
	char *name ;
	char *text ;	/* Contents of file */
	int a, u, o;	/* Sizes of buffer  */
	int nlines ;	/* Number of lines  */
	char **alines ;	/* Offsets of lines */
} SFILE ;
static SFILE sfile[MAXFILES] ;

static int nfiles;	/* Number of files  */
static int (*fdiff)() ;	/* Comparison funct */

static char ov  = 0;	/* Verbose Option  */
static char optd = 1;	/* -f option	   */

typedef int (*SORT_FCT)() ;

/* Macros with their interpretation; default = text */
char *tex_symb[] = {
  "AA", "A",
  "aa", "a",
  "c", "", 	/* Accent */
  "v", "", 	/* Accent */
  "deg", "", 	/* Accent */
  "&", "&", 
  "$", "$", 
  (char *)0
} ;

/*==================================================================
		General Utility Routines
 *==================================================================*/

#ifdef MISS_strncasecmp
static int strncasecmp(s1, s2, len)
/*++++++++++++++++
.PURPOSE  Compare two strings, case insensitive
.RETURNS  The difference
-----------------*/
  char *s1 ;	/* IN: First string 	*/
  char *s2 ;	/* IN: Second string 	*/
  int len ;	/* IN: Length of Comparison */
{
  char *e1;
  int diff ;
    e1 = s1 + len ;
    diff = 0; 
    while ((diff == 0) && (s1 < e1)) 
	diff = toupper(*s1) - toupper(*s2) ;
    return(diff) ;
}
#endif

#if 0
static char *strdup2(s1, s2)
/*++++++++++++++++
.PURPOSE  Create a new string by concatenation
.RETURNS  The concatenated (new) string
-----------------*/
  char *s1 ;	/* IN: First string 	*/
  char *s2 ;	/* IN: Second string 	*/
{
  char *s; int len ;
    len = strlen(s1) + 1 ;
    if (s2) len += strlen(s2) ;
    if ((s = malloc(len))) {
	strcpy(s, s1);
	if (s2) strcat(s, s2) ;
    }
    return(s) ;
}

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

static int strbloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character backward
.RETURNS  Index of located char / -1 if not found
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for (s=text+strlen(text); --s >= text; )  if (*s == c)    break;
    return(s-text);
}
#endif

static char *lookup(dic, term)
/*++++++++++++++++
.PURPOSE  Find equivalence in a Dictionary
.RETURNS  The equivqlence : NULL
-----------------*/
  char **dic ;	/* IN: The Dictionary	*/
  char *term ;	/* IN: The term lookup	*/
{
  char *p, *t ;
    while (*dic) {
	for (p = *dic, t = term; isalpha(*p) && isalpha(*t) && (*p == *t); 
	    p++, t++) ;
	if ((!isalpha(*p)) && (!isalpha(*t))) return(dic[1]) ;
        dic += 2 ;
    }
    return ((char *)0) ;
}

/*==================================================================
		Comparison (difference) Functions
 *==================================================================*/
static int dif_d(s1, s2, len)
/*++++++++++++++++
.PURPOSE  Compare two strings, dictionary order, maximum len bytes
.RETURNS  The difference (0 when matching)
.REMARKS  len applies to s1
-----------------*/
  char *s1;	/* IN: First string	*/
  char *s2;	/* IN: Second string	*/
  int len ;	/* IN: Max. s1 length	*/
{
  char *e1;
  int diff ;
    e1 = s1 + len ;
    diff = 0; 
    while ((diff == 0) && (s1 < e1))  {
	if (ispunct(*s1)) { s1++; continue; }
	if (ispunct(*s2)) { s2++; continue; }
	diff = toupper(*s1) - toupper(*s2) ;
	if (isspace(*s1)) while (isspace(*s1)) s1++ ;
	else s1++;
	if (isspace(*s2)) while (isspace(*s2)) s2++ ;
	else s2++;
    }
    return(diff) ;
}

static int skip_tex(str, equiv)
/*++++++++++++++++
.PURPOSE  Compare two strings, latex order, case inensitive, maximum len bytes
.RETURNS  The length of macro to skip
-----------------*/
  char *str;	/* IN: string to check, starting by \\	*/
  char *equiv;	/* OUT: Equivalence of Input */
{
  char *s, *s0, *eq ; int m ;
    s = str ;
    equiv[0] = 0 ;
    if (*s == '\\') s++ ;
    if (isalpha(*s)) {
	s0 = s ;
	while (isalpha(*s)) s++ ;
	if ((eq = lookup(tex_symb, s0))) strcpy(equiv, eq) ;
	else { 
	    strncpy(equiv, s0, m = s - s0) ; 
	    equiv[m] = 0; 
	}
	while (*s == ' ') s++ ;
    }
    else /* e.g. \'e */ s++ ;
    return(s-str) ;
}

static int dif_tex(s1, s2, len)
/*++++++++++++++++
.PURPOSE  Compare two strings, latex order, case inensitive, maximum len bytes
.RETURNS  The difference (0 when matching)
.REMARKS  len applies to s1
-----------------*/
  char *s1;	/* IN: First string	*/
  char *s2;	/* IN: Second string	*/
  int len ;	/* IN: Max.s1 length	*/
{
  static char *n1, *n2 ;
  static int recu ;
  char *e1, *sn1, *sn2, eq1[20], eq2[20];
  int diff ;
    e1 = s1 + len ;
    diff = 0; 
    recu++;
    if (ov) {
    printf("...'") ;
    for (sn1 = s1; isgraph(*sn1); sn1++) putchar(*sn1) ;
    if (*sn1 == '\t') { putchar('='), sn1++; while(isgraph(*sn1)) putchar(*sn1), sn1++; }
    printf("' - '") ;
    for (sn1 = s2; isgraph(*sn1); sn1++) putchar(*sn1) ;
    if (*sn1 == '\t') { putchar('='), sn1++; while(isgraph(*sn1)) putchar(*sn1), sn1++; }
    printf("' =>") ;
    }
    while ((diff == 0) && (s1 < e1)) {
	if (*s1 == '{') { s1++; continue; }
	else if (*s1 == '}') { s1++; continue; }
	else if (*s1 == '\\') { 
	    s1 += skip_tex(s1, eq1); 
	    if (eq1[0]) { 
		sn1 = n1; sn2 = n2 ;
	        diff = dif_tex(eq1, s2, strlen(eq1)) ;
		s2 = n2 ;
		n1 = sn1 ; n2 = sn2 ;
	    }
	    continue; 
	}
	else if (optd && ispunct(*s1)) { s1++; continue; }
	if (*s2 == '{') { s2++; continue; }
	else if (*s2 == '}') { s2++; continue; }
	else if (*s2 == '\\') { 
	    s2 += skip_tex(s2, eq2); 
	    if (eq2[0]) {
		sn1 = n1; sn2 = n2 ;
	        diff = dif_tex(s1, eq2, strlen(eq2)) ;
		s1 = n1 ;
		n1 = sn1 ; n2 = sn2 ;
	    }
	    continue; 
	}
	else if (optd && ispunct(*s2)) { s2++; continue; }
	diff = toupper(*s1) - toupper(*s2) ;
	if (isspace(*s1)) while (isspace(*s1)) s1++ ;
	else s1++ ;
	if (isspace(*s2)) while (isspace(*s2)) s2++ ;
	else s2++ ;
    }
    n1 = s1 ; n2 = s2 ;	/* Save the new new position */
    if (ov) printf("%d\n", diff) ;
    recu-- ;
    return(diff) ;
}

static int sort_fct(a1, a2)
/*++++++++++++++++
.PURPOSE  Function required for qsort
.RETURNS  Difference
-----------------*/
  char **a1 ;	/* IN: First address	*/
  char **a2 ;	/* IN: Second address	*/
{
    return ((*fdiff)(*a1, *a2, 256)) ;
}

/*==================================================================
		File's Extracts
 *==================================================================*/

static int LoadFile(sf)
/*++++++++++++++++
.PURPOSE  Load a file into memory
.RETURNS  -1 = error
.REMARKS  
-----------------*/
  SFILE *sf ;	/* MOD: File to load   */
{
  int len, i, n ;
  char *p, *e, **a ;
    
    i = open(sf->name, O_BINARY) ;
    if (i <= 0) {
	fprintf(stderr, "****") ;
	perror(sf->name) ;
	return(i) ;
    }

    sf->u = lseek(i, 0L, 2); lseek(i, 0L, 0);
    sf->a = 1+(sf->u|15) ;
    sf->text = malloc(sf->a) ;
    len = read(i, sf->text, sf->u) ;
    if (len != sf->u) {
	fprintf(stderr, "****Error reading %s (%d instead of %d bytes\n", 
	  sf->name, len, sf->u) ;
	return(-1) ;
    }
    if (len < 100) fprintf(stderr, 
	"++++File %s is only %d bytes\n", sf->name, len) ;
    sf->u += 1 ;
    sf->text[sf->u] = 0 ;
    close(i);

    /* Count number of lines */
    for (p = sf->text, e = p+sf->u, n=0; p < e; p++) {
	if (*p == '\n') n++ ;
    }
    sf->nlines = n ;
    sf->alines = a = (char **)malloc((n+2)*sizeof(char *)) ;

    /* Set the pointers */
    *(a++) = p = sf->text ;
    --e ;			/* To stop before the last \n */
    while (p < e) {
	if (*(p++) != '\n') continue ;
	*(a++) = p ;
    }
    *a = (char *)0 ;
    
    return(0);
}

/*==================================================================*/

int main(int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Main
.RETURNS  to OS
.REMARKS  
-----------------*/
{
  char *p;
  FILE *out ;
  char *fout;	/* Output */
  char **a ;
  int  i;

    fdiff = strncmp ;	/* Default Comparison Function 	*/
    fout = (char *)0 ;

    /* Take care of Options */
    while (--argc > 0) {
	p = *++argv;
	if (*p == '-') switch(p[1]) {	/* Option 	*/
	    case 'h':
		printf("%s%s", usage, help) ;
		exit(0) ;
	    case 'd':			/* Dictionary Order */
		if (fdiff == strncmp) fdiff = dif_d ;
		optd = 1 ;
		continue ;
	    case 'T':			/* Dictionary Order */
		fdiff = dif_tex ;
		continue ;
	    case 'f':			/* Case-insensitive */
		fdiff = strncasecmp ;
		continue ;
	    case 'o':			/* Output File	*/
		fout = *++argv ;
		--argc ;
		continue ;
	    case 'v': 			/* Verbose	*/
		ov = 1;
		continue;
	    default:
	    bad_option:
	    	fprintf(stderr, "*** bsort: Invalid Option: %s\n", p);
		fprintf(stderr, "%s", usage) ;
	    	exit(1) ;
	}
	if (nfiles >= MAXFILES) {
	    fprintf(stderr, "*** bsort: Too many (%d) files: %s\n", 
	    	nfiles, p);
	    fprintf(stderr, "%s", usage) ;
	    exit(1) ;
	}
	else sfile[nfiles].name = p ;
	sfile[nfiles].name = strdup(p) ;
	if (LoadFile(&sfile[nfiles]) < 0) exit(1) ;
	nfiles++;
    }

    /* Check everything is here */
    if (!nfiles) {
	fprintf(stderr, "*** bsort: Missing file(s) arguments\n");
	fprintf(stderr, "%s", usage) ;
	exit(1) ;
    }

    /* Sort each file */
    for (i=0; (i<nfiles);  i++) {
	qsort(sfile[i].alines, sfile[i].nlines, sizeof(char *), 
	    (SORT_FCT)sort_fct) ;
    }

    /* Rewrite the Output */
    if (fout) out = fopen(fout, "w") ;
    else out = stdout ;
    if (!out) { fprintf(stderr, "****") ; perror(fout); exit(1); }
    for (i=0; i<nfiles;  i++) {
	for (a=sfile[i].alines; *a; a++) { 
	    for (p = *a; *p ; p++) { fputc(*p, out) ; if (*p == '\n') break ; }
	}
    }
    if (out != stdout) fclose(out) ;
    exit(0);
}
